# -*- coding: utf8 -*-

# Copyright 2011-2014 Facundo Batista
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# For further info, check  https://launchpad.net/encuentro

"""Some functions to deal with network and Encuentro site."""

import logging
import os
import sys
import time
import urllib
import urllib2

from threading import Thread, Event
from Queue import Queue, Empty

import bs4
import defer
import requests

if __name__ == '__main__':
    # special import before any other imports to configure GUI to use API 2; we
    # normally don't need to do this *here*, just a support for run
    # this as a script, for testing/development purpuses
    import sip
    for n in "QDate QDateTime QString QTextStream QTime QUrl QVariant".split():
        sip.setapi(n, 2)   # API v2 FTW!

from PyQt4 import QtNetwork, QtCore

from encuentro import multiplatform
from encuentro.config import config

AUTH_URL = "http://registro.educ.ar/cuentas/ServicioLogin/index"
CHUNK = 16 * 1024
MB = 1024 ** 2

BAD_LOGIN_TEXT = "Ingreso de usuario"

DONE_TOKEN = "I positively assure that the download is finished (?)"

logger = logging.getLogger('encuentro.network')


class BadCredentialsError(Exception):
    """Problems with user and/or password."""


class EncuentroError(Exception):
    """Generic problem working with the Encuentro web site."""
    def __init__(self, message, original_exception=None):
        self.orig_exc = original_exception
        super(EncuentroError, self).__init__(message)


class CancelledError(Exception):
    """The download was cancelled."""


class MiBrowser(Thread):
    """Threaded browser to do the download."""

    # we *are* calling parent's init; pylint: disable=W0231
    def __init__(self, parent, authuser, authpass, url,
                 fname, output_queue, must_quit):
        self.parent = parent
        self.authinfo = authuser, authpass
        self.url = url
        self.fname = fname
        self.output_queue = output_queue
        self.must_quit = must_quit
        super(MiBrowser, self).__init__()

    def _get_download_content(self):
        """Get the content handler to download."""
        # log in
        logger.debug("Browser download, authenticating")
        usr, psw = self.authinfo
        get_data = dict(
            servicio=self.parent.service,
            continuar=urllib.quote(self.url),
        )
        complete_auth_url = AUTH_URL + "?" + urllib.urlencode(get_data)
        post_data = dict(
            login_user_name=usr,
            login_user_password=psw,
            r=complete_auth_url,
        )
        sess = requests.Session()
        sess.post(complete_auth_url, post_data)

        # get page with useful link
        logger.debug("Browser download, getting html")
        html = sess.get(complete_auth_url).content
        if BAD_LOGIN_TEXT in html:
            logger.error("Wrong user or password sent")
            raise BadCredentialsError()
        logger.debug("Browser download, got html len %d", len(html))

        # download from the new url
        html = sess.get(self.url).content
        soup = bs4.BeautifulSoup(html)
        new_url = soup.find(attrs={'class': 'descargas panel row'}).find('a')['href']
        logger.debug("Opening final url %r", new_url)
        content = urllib2.urlopen(new_url)
        try:
            filesize = int(content.headers['content-length'])
        except KeyError:
            logger.debug("No content information")
        else:
            logger.debug("Got content! filesize: %d", filesize)
            return content, filesize

        # ok, we don't know what happened :(
        logger.error("Unknown error while browsing Encuentro: %r", html)
        raise EncuentroError("Unknown problem when getting download link")

    def run(self):
        """Do the heavy work."""
        # open the url and send the content
        logger.debug("Browser opening url %s", self.url)
        try:
            content, filesize = self._get_download_content()
        except Exception as err:
            self.output_queue.put(err)
            return

        aout = open(self.fname, "wb")
        tot = 0
        size_mb = filesize / (1024.0 ** 2)
        while not self.must_quit.is_set():
            r = content.read(CHUNK)
            if r == "":
                break
            aout.write(r)
            tot += len(r)
            m = "%.1f%% (de %d MB)" % (tot * 100.0 / filesize, size_mb)
            self.output_queue.put(m)
        content.close()
        self.output_queue.put(DONE_TOKEN)


class DeferredQueue(Queue):
    """A Queue with a deferred get."""

    _call_period = 500

    def deferred_get(self):
        """Return a deferred that is triggered when data."""
        d = defer.Deferred()
        attempts = [None] * 6

        def check():
            """Check if we have data and transmit it."""
            try:
                data = self.get(block=False)
            except Empty:
                # no data, check again later, unless we had too many attempts
                attempts.pop()
                if attempts:
                    QtCore.QTimer.singleShot(self._call_period, check)
                else:
                    # finish without data, for external loop to do checks
                    d.callback(None)
            else:
                # have some data, let's check if there's more
                all_data = [data]
                try:
                    while True:
                        all_data.append(self.get(block=False))
                except Empty:
                    # we're done!
                    d.callback(all_data)

        QtCore.QTimer.singleShot(self._call_period, check)
        return d


class BaseDownloader(object):
    """Base episode downloader."""

    def shutdown(self):
        """Quit the download."""
        return self._shutdown()

    def cancel(self):
        """Cancel a download."""
        return self._cancel()

    def _setup_target(self, channel, section, season, title, extension):
        """Set up the target file to download."""
        # build where to save it
        downloaddir = config.get('downloaddir', '')
        channel = multiplatform.sanitize(channel)
        section = multiplatform.sanitize(section)
        title = multiplatform.sanitize(title)

        if season is not None:
            season = multiplatform.sanitize(season)
            fname = os.path.join(downloaddir, channel, section,
                                 season, title + extension)
        else:
            fname = os.path.join(downloaddir, channel, section,
                                 title + extension)

        # if the directory doesn't exist, create it
        dirsecc = os.path.dirname(fname)
        if not os.path.exists(dirsecc):
            os.makedirs(dirsecc)

        tempf = fname + str(time.time())
        return fname, tempf

    def download(self, channel, section, season, title, url, cb_progress):
        """Download an episode."""
        return self._download(channel, section, season,
                              title, url, cb_progress)


class AuthenticatedDownloader(BaseDownloader):
    """Episode downloader for Conectar site."""

    def __init__(self):
        super(AuthenticatedDownloader, self).__init__()
        self._prev_progress = None
        self.browser_quit = set()
        self.cancelled = False
        logger.info("Conectar downloader inited")

    def _shutdown(self):
        """Quit the download."""
        for bquit in self.browser_quit:
            bquit.set()
        logger.info("Conectar downloader shutdown finished")

    def _cancel(self):
        """Cancel a download."""
        self.cancelled = True
        logger.info("Conectar downloader cancelled")

    @defer.inline_callbacks
    def _download(self, canal, seccion, season, titulo, url, cb_progress):
        """Download an episode to disk."""
        self.cancelled = False

        # levantamos el browser
        qinput = DeferredQueue()
        bquit = Event()
        self.browser_quit.add(bquit)
        authuser = config.get('user', '')
        authpass = config.get('password', '')

        # build where to save it
        fname, tempf = self._setup_target(canal, seccion, season,
                                          titulo, u".avi")
        logger.debug("Downloading to temporal file %r", tempf)

        logger.info("Download episode %r: browser started", url)
        brow = MiBrowser(self, authuser, authpass, url, tempf, qinput, bquit)
        brow.start()

        # loop reading until finished
        self._prev_progress = None

        logger.info("Downloader started receiving bytes")
        while True:
            # get all data and just use the last item
            payload = yield qinput.deferred_get()
            if self.cancelled:
                logger.debug("Cancelled! Quit browser, wait, and clean.")
                bquit.set()
                yield qinput.deferred_get()
                if os.path.exists(tempf):
                    os.remove(tempf)
                logger.debug("Cancelled! Cleaned up.")
                raise CancelledError()

            # special situations
            if payload is None:
                # no data, let's try again
                continue
            data = payload[-1]
            if isinstance(data, Exception):
                raise data
            if data == DONE_TOKEN:
                break

            # actualizamos si hay algo nuevo
            if data != self._prev_progress:
                cb_progress(data)
                self._prev_progress = data

        # movemos al nombre correcto y terminamos
        logger.info("Downloading done, renaming temp to %r", fname)
        os.rename(tempf, fname)
        self.browser_quit.remove(bquit)
        defer.return_value(fname)


class ConectarDownloader(AuthenticatedDownloader):
    """Episode downloader for Conectar site."""
    service = 'conectate'


class EncuentroDownloader(AuthenticatedDownloader):
    """Episode downloader for Conectar site."""
    service = 'encuentro'


class _GenericDownloader(BaseDownloader):
    """Episode downloader for a generic site that works with urllib2."""

    headers = {
        'User-Agent': 'Mozilla/5.0',
        'Accept': '*/*',
    }
    manager = QtNetwork.QNetworkAccessManager()
    file_extension = None  # to be overwritten by class child

    def __init__(self):
        super(_GenericDownloader, self).__init__()
        self._prev_progress = None
        self.downloader_deferred = None
        logger.info("Generic downloader inited")

    def _shutdown(self):
        """Quit the download."""
        logger.info("Generic downloader shutdown finished")

    def _cancel(self):
        """Cancel a download."""
        if self.downloader_deferred is not None:
            logger.info("Generic downloader cancelled")
            exc = CancelledError("Cancelled by user")
            self.downloader_deferred.errback(exc)

    @defer.inline_callbacks
    def _download(self, canal, seccion, season, titulo, url, cb_progress):
        """Download an episode to disk."""
        url = str(url)
        logger.info("Download episode %r", url)

        # build where to save it
        fname, tempf = self._setup_target(canal, seccion,
                                          season, titulo, self.file_extension)
        logger.debug("Downloading to temporal file %r", tempf)
        fh = open(tempf, "wb")

        def report(dloaded, total):
            """Report download."""
            if total == -1:
                m = "%d MB" % (dloaded // MB,)
            else:
                size_mb = total // MB
                perc = dloaded * 100.0 / total
                m = "%.1f%% (de %d MB)" % (perc, size_mb)
            if m != self._prev_progress:
                cb_progress(m)
                self._prev_progress = m

        def save():
            """Save available bytes to disk."""
            data = req.read(req.bytesAvailable())
            fh.write(data)

        request = QtNetwork.QNetworkRequest()
        request.setUrl(QtCore.QUrl(url))
        for hk, hv in self.headers.items():
            request.setRawHeader(hk, hv)

        def end_ok():
            """Finish Ok politely the deferred."""
            if not self.downloader_deferred.called:
                self.downloader_deferred.callback(True)

        def end_fail(exc):
            """Finish in error politely the deferred."""
            if not self.downloader_deferred.called:
                self.downloader_deferred.errback(exc)

        deferred = self.downloader_deferred = defer.Deferred()
        req = self.manager.get(request)
        req.downloadProgress.connect(report)
        req.error.connect(end_fail)
        req.readyRead.connect(save)
        req.finished.connect(end_ok)

        try:
            yield deferred
        except Exception as err:
            logger.debug("Exception when waiting deferred: %s (request "
                         "finished? %s)", err, req.isFinished())
            if not req.isFinished():
                logger.debug("Aborting QNetworkReply")
                req.abort()
            raise
        finally:
            fh.close()

        # rename to final name and end
        logger.info("Downloading done, renaming temp to %r", fname)
        os.rename(tempf, fname)
        defer.return_value(fname)


class GenericVideoDownloader(_GenericDownloader):
    """Generic downloaded that saves video."""
    file_extension = u".mp4"


class GenericAudioDownloader(_GenericDownloader):
    """Generic downloaded that saves audio."""
    file_extension = u".mp3"


# this is the entry point to get the downloaders for each type
all_downloaders = {
    'encuentro': EncuentroDownloader,
    'conectar': ConectarDownloader,
    'generic': GenericVideoDownloader,
    'dqsv': GenericAudioDownloader,
}


if __name__ == "__main__":
    h = logging.StreamHandler()
    h.setLevel(logging.DEBUG)
    logger.setLevel(logging.DEBUG)
    logger.addHandler(h)

    def show(avance):
        """Show progress."""
        print "Avance:", avance

    # overwrite config for the test
    config = dict(user="lxpdvtnvrqdoa@mailinator.com",  # NOQA
                  password="descargas", downloaddir='.')

    # the three versions to test
    downloader = EncuentroDownloader()
    _url = "http://www.encuentro.gob.ar/sitios/encuentro/"\
           "Programas/ver?rec_id=120761"

#    downloader = ConectarDownloader()
#    _url = "http://www.conectate.gob.ar/sitios/conectate/"\
#           "busqueda/pakapaka?rec_id=103605"

    app = QtCore.QCoreApplication(sys.argv)
#    downloader = GenericVideoDownloader()
#    _url = "http://backend.bacua.gob.ar/video.php?v=_f9d06f72"

    @defer.inline_callbacks
    def download():
        """Download."""
        try:
            fname = yield downloader.download("test-ej-canal", "secc", "temp",
                                              "tit", _url, show)
            print "All done!", fname
        except CancelledError:
            print "--- cancelado!"
        finally:
            downloader.shutdown()
            app.exit()
    download()
    sys.exit(app.exec_())



# -*- coding: UTF-8 -*-

# Copyright 2013-2014 Facundo Batista
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# For further info, check  https://launchpad.net/encuentro

"""Central panels in the main window, the content part of all the interface."""

import logging
import operator

from PyQt4.QtGui import (
    QAbstractItemView,
    QAbstractTextDocumentLayout,
    QApplication,
    QBrush,
    QColor,
    QHBoxLayout,
    QImage,
    QLabel,
    QMenu,
    QPixmap,
    QPushButton,
    QStyle,
    QStyleOptionViewItemV4,
    QStyledItemDelegate,
    QTextDocument,
    QTextEdit,
    QTreeWidgetItem,
    QVBoxLayout,
    QWidget,
)
from PyQt4.QtCore import Qt, QSize, QAbstractTableModel

from encuentro import data, image
from encuentro.config import config, signal
from encuentro.data import Status
from encuentro.ui import remembering
from encuentro.ui.throbber import Throbber

logger = logging.getLogger("encuentro.centralpanel")


class DownloadsWidget(remembering.RememberingTreeWidget):
    """The downloads queue."""

    def __init__(self, episodes_widget):
        self.episodes_widget = episodes_widget
        super(DownloadsWidget, self).__init__('downloads')
        signal.register(self.save_state)

        _headers = (u"Descargando...", u"Estado")
        self.setColumnCount(len(_headers))
        self.setHeaderLabels(_headers)

        self.queue = []
        self.current = -1
        self.downloading = False

        # connect the signals
        self.clicked.connect(self.on_signal_clicked)

    def on_signal_clicked(self, _):
        """The view was clicked."""
        item = self.currentItem()
        self.episodes_widget.show_episode(item.episode_id)

    def append(self, episode):
        """Append an episode to the downloads list."""
        # add to the list in the GUI
        item = QTreeWidgetItem((episode.composed_title, u"Encolado"))
        item.episode_id = episode.episode_id
        self.queue.append((episode, item))
        self.addTopLevelItem(item)
        self.setCurrentItem(item)

        # fix episode state
        episode.state = Status.waiting

    def prepare(self):
        """Set up everything for next download."""
        self.downloading = True
        self.current += 1
        episode, _ = self.queue[self.current]
        episode.state = Status.downloading
        return episode

    def start(self):
        """Download started."""
        episode, item = self.queue[self.current]
        item.setText(1, u"Comenzando")
        episode.state = Status.downloading

    def progress(self, progress):
        """Advance the progress indicator."""
        _, item = self.queue[self.current]
        item.setText(1, u"Descargando: %s" % progress)

    def end(self, error=None):
        """Mark episode as downloaded."""
        episode, item = self.queue[self.current]
        if error is None:
            # downloaded OK
            gui_msg = u"Terminado ok"
            end_state = Status.downloaded
        else:
            # something bad happened
            gui_msg = unicode(error)
            end_state = Status.none
        item.setText(1, gui_msg)
        item.setDisabled(True)
        episode.state = end_state
        self.episodes_widget.refresh(episode.episode_id)
        self.downloading = False

    def cancel(self):
        """The download is being cancelled."""
        episode, item = self.queue[self.current]
        item.setText(1, u"Cancelado")
        episode.state = Status.none

    def unqueue(self, episode):
        """Remove the indicated episode from the queue."""
        episode.state = Status.none

        # search for the item, adjust the queue and remove it from the widget
        for pos, (queued_episode, item) in enumerate(self.queue):
            if queued_episode.episode_id == episode.episode_id:
                break
        else:
            raise ValueError(
                "Couldn't find episode to unqueue: " + str(episode))
        del self.queue[pos]
        self.takeTopLevelItem(pos)

        # as we removed an item, the cursor goes to other, fix the rest of
        # the interface
        item = self.currentItem()
        self.episodes_widget.show_episode(item.episode_id)

    def pending(self):
        """Return the pending downloads quantity (including current)."""
        # remaining after current one
        q = len(self.queue) - self.current - 1
        # if we're still downloading current one, add it to the count
        if self.downloading:
            q += 1
        return q

    def save_state(self):
        """Save state for pending downloads."""
        p = self.pending()
        if p > 0:
            pending_ids = [e.episode_id for e, _ in self.queue[-p:]]
        else:
            pending_ids = []
        config[config.SYSTEM]['pending_ids'] = pending_ids

    def load_pending(self):
        """Queue the pending downloads."""
        loaded_pending_ids = config[config.SYSTEM].get('pending_ids', [])

        for episode_id in loaded_pending_ids:
            main_window = self.episodes_widget.main_window
            try:
                episode = main_window.programs_data[episode_id]
            except KeyError:
                logger.debug("Tried to load pending %r, didn't find it",
                             episode_id)
            else:
                main_window.queue_download(episode)


class HTMLDelegate(QStyledItemDelegate):
    """Custom delegate so the QTreeWidget can do HTML.

    This is an adaptation of a post here:

        http://stackoverflow.com/questions/10924175/how-do-i-use-a-
            qstyleditemdelegate-to-paint-only-the-background-without-coverin

    We only need to do background highlighting, so probably this will be
    trimmed as much as possible for performance reasons.

    Also, we only do HTML for one column, the rest is delegated to parent.
    """
    def __init__(self, parent, html_column):
        self._html_column = html_column
        QStyledItemDelegate.__init__(self, parent)

    def paint(self, painter, option, index):
        """Render the delegate for the item."""
        if index.column() != self._html_column:
            return QStyledItemDelegate.paint(self, painter, option, index)

        options = QStyleOptionViewItemV4(option)
        self.initStyleOption(options, index)

        if options.widget is None:
            style = QApplication.style()
        else:
            style = options.widget.style()

        doc = QTextDocument()
        doc.setHtml(options.text)

        options.text = ""
        style.drawControl(QStyle.CE_ItemViewItem, options, painter)

        ctx = QAbstractTextDocumentLayout.PaintContext()

        textRect = style.subElementRect(QStyle.SE_ItemViewItemText, options)
        painter.save()
        painter.translate(textRect.topLeft())
        painter.setClipRect(textRect.translated(-textRect.topLeft()))
        doc.documentLayout().draw(painter, ctx)
        painter.restore()

    def sizeHint(self, option, index):
        """Calculate the needed size."""
        options = QStyleOptionViewItemV4(option)
        self.initStyleOption(options, index)

        doc = QTextDocument()
        doc.setHtml(options.text)
        doc.setTextWidth(options.rect.width())
        return QSize(doc.idealWidth(), doc.size().height())


class EpisodesWidgetModel(QAbstractTableModel):
    """The model for the episodes widget."""

    _col_getters = [
        operator.attrgetter('channel'),
        operator.attrgetter('section'),
        operator.attrgetter('filtered_title'),
        operator.attrgetter('duration'),
    ]
    _headers = (u"Canal", u"Sección", u"Título", u"Duración [min]")

    def __init__(self, main_window):
        super(EpisodesWidgetModel, self).__init__()
        self.main_window = main_window
        self._order_column = self._order_direction = 0
        self._filter_text = ''
        self._filter_only_downloaded = False
        self.episodes, self.pos_map = self._load_episodes()

    def _load_episodes(self):
        """Fill episodes own data."""
        # prepare sorting parameters
        is_reversed = self._order_direction == Qt.DescendingOrder
        order_key = self._col_getters[self._order_column]

        # get all episodes, apply filters
        text = data.prepare_to_filter(self._filter_text)
        episodes = []
        for ep in self.main_window.programs_data.values():
            params = ep.filter_params(text, self._filter_only_downloaded)
            if params is None:
                # filtered out
                continue

            pos1, pos2 = params
            if pos1 == pos2:
                # no highlighting
                ep.filtered_title = ep.composed_title
            else:
                # filtering by text, so highlight
                t = ep.composed_title
                ep.filtered_title = u'%s<span style="background-color:yellow">%s</span>%s' % (
                    t[:pos1], t[pos1:pos2], t[pos2:])
            episodes.append(ep)

        # episodes si the data to show in the table, with some extra columns (hidden),
        # pos_map is a mapping to know in which position an episode is from its id
        episodes = sorted(episodes, key=order_key, reverse=is_reversed)
        pos_map = {ep.episode_id: i for i, ep in enumerate(episodes)}
        return episodes, pos_map

    def reload_episodes(self):
        """Reload all episodes."""
        self.layoutAboutToBeChanged.emit()
        self.episodes, self.pos_map = self._load_episodes()
        self.layoutChanged.emit()

    def rowCount(self, parent):
        """Row count."""
        return len(self.episodes)

    def columnCount(self, parent):
        """Column count."""
        return len(self._headers)

    def data(self, index, role):
        """Return content text and format."""
        if role == Qt.DisplayRole:
            row = index.row()
            col = index.column()
            ep = self.episodes[row]
            data = self._col_getters[col](ep)
            return data

        if role == Qt.TextAlignmentRole:
            col = index.column()
            if col == 3:
                return Qt.AlignRight

        if role == Qt.BackgroundRole:
            row = index.row()
            ep = self.episodes[row]
            if ep.state == Status.downloaded:
                bground = QBrush(QColor("light green"))
                return bground

    def headerData(self, section, orientation, role):
        """Return the headers."""
        if role == Qt.DisplayRole and orientation == Qt.Horizontal:
            return self._headers[section]

    def refresh(self, episode_id):
        """Refresh the view of an episode."""
        row = self.pos_map[episode_id]
        index_from = self.index(row, 0)
        index_to = self.index(row, len(self._headers) - 1)
        self.dataChanged.emit(index_from, index_to)

    def sort(self, n_col, order):
        """Sort data by given column."""
        self._order_column = n_col
        self._order_direction = order
        self.reload_episodes()

    def set_filter(self, text, only_downloaded):
        """Apply a filter to the episodes list."""
        self._filter_text = text
        self._filter_only_downloaded = only_downloaded
        self.reload_episodes()


class EpisodesWidgetView(remembering.RememberingTableView):
    """The list of episodes info."""

    _title_column = 2

    def __init__(self, main_window, episode_info):
        self.main_window = main_window
        self.episode_info = episode_info
        super(EpisodesWidgetView, self).__init__('episodes')
        self._model = EpisodesWidgetModel(main_window)
        self.setModel(self._model)
        self.setMinimumSize(600, 300)
        self.setItemDelegate(HTMLDelegate(self, self._title_column))

        # hide the vertical header at the left of the table and configure top header
        self.verticalHeader().hide()
        header = self.horizontalHeader()
        header.setStretchLastSection(False)
        header.setResizeMode(2, header.Stretch)
        header.sortIndicatorChanged.connect(self._model.sort)

        # other behaviour configs
        self.setSelectionBehavior(QAbstractItemView.SelectRows)  # whole row selected instead cell
        self.setSelectionMode(QAbstractItemView.ExtendedSelection)  # full fledged selection
        self.setSortingEnabled(True)  # enable sorting

        # connect the signals
        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self.on_right_button)
        self.doubleClicked.connect(self.on_activate)
        self.activated.connect(self.on_activate)
        sm = self.selectionModel()
        sm.selectionChanged.connect(self.on_change)

    def show_episode(self, episode_id):
        """Show the row for the requested episode."""
        row = self._model.pos_map[episode_id]
        index = self._model.index(row, 0)
        self.scrollTo(index)
        self._adjust_gui(episode_id)

    def selected_items(self):
        """Return the episode ids of the selected items."""
        sm = self.selectionModel()
        indexes = sm.selectedRows()
        return [self._model.episodes[ind.row()].episode_id for ind in indexes]

    def on_change(self, delta_selection_pos, delta_selection_neg):
        """The view was clicked."""
        sm = self.selectionModel()
        indexes = sm.selectedRows()
        for idx in indexes:
            row = idx.row()
        if len(indexes) == 1:
            row = indexes[0].row()
            episode_id = self._model.episodes[row].episode_id
            self._adjust_gui(episode_id)
        elif len(indexes) == 0:
            # nothing selected
            self.episode_info.clear(u"Seleccioná un programa para ver aquí la info.")
            self.main_window.check_download_play_buttons()
        else:
            # multiple selection
            self.episode_info.clear(u"Seleccioná sólo un programa para ver aquí la info.")
            self.main_window.check_download_play_buttons()

    def on_activate(self, index):
        """Double click and enter on a row."""
        row = index.row()
        episode_id = self._model.episodes[row].episode_id
        episode = self.main_window.programs_data[episode_id]
        logger.debug("Doubleclick/Enter in %s", episode)
        if episode.state == Status.downloaded:
            self.main_window.play_episode(episode)
        elif episode.state == Status.none:
            if self.main_window.have_config():
                self.main_window.queue_download(episode)
            else:
                logger.debug("Not starting download because no config.")
                t = u"No se puede arrancar una descarga porque la configuración está incompleta."
                self.main_window.show_message(u'Falta configuración', t)

    def _adjust_gui(self, episode_id):
        """Adjust the rest of the GUI for this episode."""
        episode = self.main_window.programs_data[episode_id]
        logger.debug("Showing episode: %s", episode)

        # adjust the rest of the GUI
        self.episode_info.update(episode)
        self.main_window.check_download_play_buttons()

        # make the view to select this episode
        row = self._model.pos_map[episode_id]
        self.selectRow(row)

    def on_right_button(self, point):
        """Right button was pressed, build a menu."""
        index = self.indexAt(point)
        row = index.row()
        episode_id = self._model.episodes[row].episode_id
        self._adjust_gui(episode_id)
        episode = self.main_window.programs_data[episode_id]
        menu = QMenu()
        mw = self.main_window
        act_play = menu.addAction(u"&Reproducir", lambda: mw.play_episode(episode))
        act_cancel = menu.addAction(u"&Cancelar descarga", lambda: mw.cancel_download(episode))
        act_download = menu.addAction(u"&Descargar", lambda: mw.queue_download(episode))

        # set menu options according status
        state = episode.state
        if state == Status.downloaded:
            act_play.setEnabled(True)
            act_cancel.setEnabled(False)
            act_download.setEnabled(False)
        elif state == Status.downloading or state == Status.waiting:
            act_play.setEnabled(False)
            act_cancel.setEnabled(True)
            act_download.setEnabled(False)
        elif state == Status.none:
            act_play.setEnabled(False)
            act_cancel.setEnabled(False)
            if self.main_window.have_config():
                act_download.setEnabled(True)
            else:
                act_download.setEnabled(False)
        menu.exec_(self.viewport().mapToGlobal(point))

    def set_filter(self, text, only_downloaded=False):
        """Apply a filter to the episodes list (just a proxy to the model)."""
        self._model.set_filter(text, only_downloaded)

    def refresh(self, episode_id):
        """Refresh the indicated episode in the view (just a proxy to the model)."""
        self._model.refresh(episode_id)

    def reload_episodes(self):
        """Reload all the episodes (just a proxy to the model)."""
        self._model.reload_episodes()


class EpisodeInfo(QWidget):
    """Show the episode at the right."""
    def __init__(self, main_window):
        self.main_window = main_window
        super(EpisodeInfo, self).__init__()

        self.current_episode = None
        layout = QVBoxLayout(self)

        # a throbber, that we don't initially show
        self.throbber = Throbber()
        layout.addWidget(self.throbber)
        self.throbber.hide()

        # the image and its getter
        self.image_episode = QLabel()
        self.image_episode.hide()
        layout.addWidget(self.image_episode, alignment=Qt.AlignCenter)
        self.get_image = image.ImageGetter(self.image_episode_loaded).get_image

        # text area
        self.text_edit = QTextEdit(
            u"Seleccionar un programa para ver aquí la info.")
        self.text_edit.setReadOnly(True)
        layout.addWidget(self.text_edit)

        # the button
        self.button = QPushButton()
        self.button.connected = False
        self.button.hide()
        layout.addWidget(self.button)

    def image_episode_loaded(self, episode_id, image_path):
        """An image has arrived, show it only if the path is correct."""
        # only set the image if the user still have the same episode selected
        if self.current_episode != episode_id:
            return

        # load the image and show it
        pixmap = QPixmap(image_path)
        self.image_episode.setPixmap(pixmap)
        self.image_episode.show()

        # hide the throbber
        self.throbber.hide()

    def clear(self, msg):
        """Clear the episode info panel."""
        self.throbber.hide()
        self.image_episode.hide()
        self.text_edit.setText(msg)
        self.button.hide()

    def update(self, episode, force_change=True):
        """Update all the episode info."""
        if not force_change:
            # if not forced, only update what is being shown if the current episode
            # is the one to be updated
            if self.current_episode != episode.episode_id:
                return

        self.current_episode = episode.episode_id

        # image
        if episode.image_data is not None:
            # have the image data already!!
            qimg = QImage.fromData(episode.image_data)
            pixmap = QPixmap.fromImage(qimg)
            self.image_episode.setPixmap(pixmap)
            self.image_episode.show()
        elif episode.image_url is not None:
            # this must be before the get_image call, as it may call
            # immediately to image_episode_loaded (showing the image and
            # hiding the throber)
            self.image_episode.hide()
            self.throbber.show()
            # now do call the get_image
            self.get_image(episode.episode_id, episode.image_url.encode('utf-8'))

        # all description
        if episode.subtitle is None:
            msg = u"<center><h3>%s</h3></center><br/><br/>%s" % (
                episode.composed_title, episode.description)
        else:
            msg = u"<center><h3>%s</h3>%s</center><br/><br/>%s" % (
                episode.composed_title, episode.subtitle, episode.description)
        self.text_edit.setHtml(msg)

        # action button
        self.button.show()
        if episode.state == data.Status.downloaded:
            label = "Reproducir"
            func = self.main_window.play_episode
            enable = True
            remove = False
        elif episode.state == data.Status.downloading:
            label = u"Cancelar descarga"
            func = self.main_window.cancel_download
            enable = True
            remove = False
        elif episode.state == data.Status.waiting:
            label = u"Sacar de la cola"
            func = self.main_window.unqueue_download
            enable = True
            remove = True
        else:
            label = u"Descargar"
            func = self.main_window.download_episode
            enable = bool(self.main_window.have_config())
            remove = False

        def _exec(func, episode, remove):
            """Execute a function on the episode and update its info."""
            func(episode)
            if not remove:
                self.update(episode)

        # set button text, disconnect if should, and connect new func
        self.button.setEnabled(enable)
        self.button.setText(label)
        if self.button.connected:
            self.button.clicked.disconnect()
        self.button.connected = True
        self.button.clicked.connect(lambda: _exec(func, episode, remove))


class BigPanel(QWidget):
    """The big panel for the main interface with user."""

    def __init__(self, main_window):
        super(BigPanel, self).__init__()
        self.main_window = main_window

        layout = QHBoxLayout(self)

        # get this before, as it be used when creating other sutff
        episode_info = EpisodeInfo(main_window)
        self.episodes = EpisodesWidgetView(main_window, episode_info)

        # split on the right
        right_split = remembering.RememberingSplitter(Qt.Vertical, 'right')
        right_split.addWidget(episode_info)
        self.downloads_widget = DownloadsWidget(self.episodes)
        right_split.addWidget(self.downloads_widget)

        # main split
        main_split = remembering.RememberingSplitter(Qt.Horizontal, 'main')
        main_split.addWidget(self.episodes)
        main_split.addWidget(right_split)
        layout.addWidget(main_split)



# -*- coding: utf8 -*-

# Copyright 2011-2013 Facundo Batista
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# For further info, check  https://launchpad.net/encuentro

"""Multiplatform code."""

import os
import re
import subprocess
import sys
import user


_basedir = os.path.abspath(os.path.dirname(os.path.dirname(
    os.path.realpath(sys.argv[0]))))

# if the base directory was determined by setup.py, fix it
# pylint: disable=W0212
if hasattr(sys, '_INSTALLED_BASE_DIR'):
    _basedir = sys._INSTALLED_BASE_DIR

# if the base directory was mangled by PyInstaller, fix it
_frozen = False
if hasattr(sys, 'frozen'):
    _basedir = sys._MEIPASS
    _frozen = True
# pylint: enable=W0212

if sys.platform == 'win32':
    # won't find this in linux; pylint: disable=F0401
    from win32com.shell import shell, shellcon
    config_dir = shell.SHGetFolderPath(0, shellcon.CSIDL_PROFILE, None, 0)
    data_dir = shell.SHGetFolderPath(0, shellcon.CSIDL_LOCAL_APPDATA, None, 0)
    cache_dir = data_dir
    del shell, shellcon
else:
    from xdg import BaseDirectory
    config_dir = BaseDirectory.xdg_config_home
    data_dir = BaseDirectory.xdg_data_home
    cache_dir = BaseDirectory.xdg_cache_home
    del BaseDirectory


def get_path(path):
    """Build an usable path for media."""
    parts = path.split("/")

    # if frozen by PyInstaller, all stuff is in the same dir
    if _frozen:
        return os.path.join(_basedir, parts[-1])

    # normal work
    return os.path.join(_basedir, *parts)


def sanitize(name):
    """Sanitize the name according to the OS."""
    if sys.platform == 'win32':
        sanit = re.sub(u'[<>:"/|?*]', '', name)
    else:
        sanit = re.sub(u'/', '', name)
    return sanit


def get_download_dir():
    """Get a the download dir for the system.

    I hope this someday will be included in the xdg library :|
    """
    try:
        cmd = ["xdg-user-dir", 'DOWNLOAD']
        proc = subprocess.Popen(cmd, stdout=subprocess.PIPE)
        base = proc.communicate()[0].strip()
    except OSError:
        base = user.home
    return os.path.join(base, 'encuentro')


def open_file(fullpath):
    """Open the file."""
    if sys.platform == 'win32':
        os.startfile(fullpath)
    else:
        subprocess.call(["/usr/bin/xdg-open", fullpath])



# -*- coding: UTF-8 -*-

# Copyright 2013 Facundo Batista
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# For further info, check  https://launchpad.net/encuentro

"""The remembering widgets."""

from PyQt4.QtCore import Qt
from PyQt4.QtGui import (
    QMainWindow,
    QSplitter,
    QTableView,
    QTreeWidget,
)

from encuentro.config import config, signal

SYSTEM = config.SYSTEM


class RememberingMainWindow(QMainWindow):
    """A MainWindow that remembers size and position."""

    def __init__(self):
        super(RememberingMainWindow, self).__init__()
        signal.register(self.save_state)
        self._name = self.__class__.__name__
        self._initted = False

    def showEvent(self, event):
        """Know when it was shown, load config."""
        if not self._initted:
            self._initted = True
            conf = config[SYSTEM].get(self._name, {})
            prv_size = conf.get('size', (800, 600))
            prv_pos = conf.get('pos', (300, 300))
            self.resize(*prv_size)
            self.move(*prv_pos)
        super(RememberingMainWindow, self).showEvent(event)

    def save_state(self):
        """Save what to remember."""
        qsize = self.size()
        size = qsize.width(), qsize.height()
        qpos = self.pos()
        pos = qpos.x(), qpos.y()
        to_save = dict(pos=pos, size=size)
        config[SYSTEM][self._name] = to_save


class RememberingSplitter(QSplitter):
    """A Splitter that remembers position."""

    def __init__(self, type_, name):
        super(RememberingSplitter, self).__init__(type_)
        signal.register(self.save_state)
        cname = self.__class__.__name__
        self._name = '-'.join((cname, name))
        self._initted = False

    def showEvent(self, event):
        """Know when it was shown, load config."""
        if not self._initted:
            self._initted = True
            sizes = config[SYSTEM].get(self._name)
            if sizes is not None:
                self.setSizes(sizes)
        super(RememberingSplitter, self).showEvent(event)

    def save_state(self):
        """Save what to remember."""
        sizes = self.sizes()
        config[SYSTEM][self._name] = sizes


class RememberingTreeWidget(QTreeWidget):
    """A TreeWidget that remembers visual stuff."""

    def __init__(self, name):
        super(RememberingTreeWidget, self).__init__()
        signal.register(self.save_state)
        cname = self.__class__.__name__
        self._name = '-'.join((cname, name))
        self._initted = False

    def showEvent(self, event):
        """Know when it was shown, load config."""
        if not self._initted:
            self._initted = True
            info = config[SYSTEM].get(self._name)
            if info is not None:
                cols_w = info['cols_w']
                for i, w in enumerate(cols_w):
                    self.setColumnWidth(i, w)
                s_enabled = info['s_enabled']
                self.setSortingEnabled(s_enabled)
                if s_enabled:
                    s_column = info['s_column']
                    s_order = info['s_order']
                    ordr = Qt.AscendingOrder if s_order else Qt.DescendingOrder
                    self.sortItems(s_column, ordr)

        super(RememberingTreeWidget, self).showEvent(event)

    def save_state(self):
        """Save what to remember."""
        cols_w = [self.columnWidth(i) for i in xrange(self.columnCount())]
        s_enabled = self.isSortingEnabled()
        s_column = self.sortColumn()
        c = self.topLevelItemCount()
        if c < 2:  # less than two records, no point in sorting
            s_order = True
        else:
            val_first = self.topLevelItem(0).text(s_column)
            val_last = self.topLevelItem(c - 1).text(s_column)
            s_order = val_first < val_last
        info = dict(cols_w=cols_w, s_enabled=s_enabled,
                    s_column=s_column, s_order=s_order)
        config[SYSTEM][self._name] = info


class RememberingTableView(QTableView):
    """A TableView that remembers visual stuff."""

    def __init__(self, name):
        super(RememberingTableView, self).__init__()
        signal.register(self.save_state)
        cname = self.__class__.__name__
        self._name = '-'.join((cname, name))
        self._initted = False

    def showEvent(self, event):
        """Know when it was shown, load config."""
        if not self._initted:
            self._initted = True
            info = config[SYSTEM].get(self._name)
            if info is not None:
                # cols width
                cols_w = info['cols_w']
                for i, w in enumerate(cols_w):
                    self.setColumnWidth(i, w)

                # sort
                sort_section = info.get('sort_section', 0)
                sort_order = info.get('sort_order', 0)
                self.sortByColumn(sort_section, sort_order)

        super(RememberingTableView, self).showEvent(event)

    def save_state(self):
        """Save what to remember."""
        # sorting info, from the header
        header = self.horizontalHeader()
        sort_section = header.sortIndicatorSection()
        sort_order = header.sortIndicatorOrder()

        # rest of values and store
        col_count = self.model().columnCount(None)
        cols_w = [self.columnWidth(i) for i in xrange(col_count)]
        info = dict(cols_w=cols_w, sort_section=sort_section, sort_order=sort_order)
        config[SYSTEM][self._name] = info



# -*- coding: utf-8 -*-

# Copyright 2013-2015 Facundo Batista
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# For further info, check  https://launchpad.net/encuentro

"""The main window."""

import logging
import os
import datetime as dt

import defer

from PyQt4.QtGui import (
    QAction,
    QCheckBox,
    QKeySequence,
    QLabel,
    QLineEdit,
    QMessageBox,
    QShortcut,
    QSizePolicy,
    QStyle,
    QWidget,
)

from encuentro import multiplatform, data, update
from encuentro.config import config, signal
from encuentro.data import Status
from encuentro.network import (
    BadCredentialsError,
    CancelledError,
    EncuentroError,
    all_downloaders,
)
from encuentro.notify import notify
from encuentro.ui import (
    central_panel,
    preferences,
    remembering,
    systray,
    wizard,
)

logger = logging.getLogger('encuentro.main')

# tooltips for buttons enabled and disabled
TTIP_PLAY_E = u'Reproducir el programa'
TTIP_PLAY_D = (
    u"Reproducir - El episodio debe estar descargado para poder verlo."
)
TTIP_DOWNLOAD_E = u'Descargar el programa de la web'
TTIP_DOWNLOAD_D = (
    u"Descargar - No se puede descargar si ya está descargado o falta "
    u"alguna configuración en el programa."
)

ABOUT_TEXT = u"""
<center>
Simple programa que permite buscar, descargar y ver<br/>
contenido del canal Encuentro y otros.<br/>
<br/>
Versión %s<br/>
<br/>
<small>Copyright 2010-2013 Facundo Batista</small><br/>
<br/>
<a href="http://encuentro.taniquetil.com.ar">
    http://encuentro.taniquetil.com.ar
</a>
</center>
"""


class MainUI(remembering.RememberingMainWindow):
    """Main UI."""

    _programs_file = os.path.join(multiplatform.data_dir, 'encuentro.data')

    def __init__(self, version, app_quit):
        super(MainUI, self).__init__()
        self.app_quit = app_quit
        self.finished = False
        self.version = version
        self.setWindowTitle('Encuentro')

        self.programs_data = data.ProgramsData(self, self._programs_file)
        self._touch_config()

        self.downloaders = {}
        for downtype, dloader_class in all_downloaders.iteritems():
            self.downloaders[downtype] = dloader_class()

        # finish all gui stuff
        self.big_panel = central_panel.BigPanel(self)
        self.episodes_list = self.big_panel.episodes
        self.episodes_download = self.big_panel.downloads_widget
        self.setCentralWidget(self.big_panel)

        # the setting of menubar should be almost in the end, because it may
        # trigger the wizard, which needs big_panel and etc.
        self.action_play = self.action_download = None
        self.filter_line = self.filter_cbox = self.needsomething_alert = None
        self._menubar()

        systray.show(self)

        if config.get('autorefresh'):
            ue = update.UpdateEpisodes(self)
            ue.background()
        else:
            # refresh data if never done before or if last
            # update was 7 days ago
            last_refresh = config.get('autorefresh_last_time')
            if last_refresh is None or (
                    dt.datetime.now() - last_refresh > dt.timedelta(7)):
                ue = update.UpdateEpisodes(self)
                ue.background()

        self.show()

        self.episodes_download.load_pending()
        logger.debug("Main UI started ok")

    def _touch_config(self):
        """Do some config processing."""
        # log the config, but without user and pass
        safecfg = config.sanitized_config()
        logger.debug("Configuration loaded: %s", safecfg)

        # we have a default for download dir
        if not config.get('downloaddir'):
            config['downloaddir'] = multiplatform.get_download_dir()

        # maybe clean some config
        if self.programs_data.reset_config_from_migration:
            config['user'] = ''
            config['password'] = ''
            config.pop('cols_width', None)
            config.pop('cols_order', None)
            config.pop('selected_row', None)

    def have_config(self):
        """Return if some config is needed."""
        return config.get('user') and config.get('password')

    def have_metadata(self):
        """Return if metadata is needed."""
        return bool(self.programs_data)

    def _menubar(self):
        """Set up the menu bar."""
        menubar = self.menuBar()

        # applications menu
        menu_appl = menubar.addMenu(u'&Aplicación')

        icon = self.style().standardIcon(QStyle.SP_BrowserReload)
        action_reload = QAction(icon, '&Refrescar', self)
        action_reload.setShortcut('Ctrl+R')
        action_reload.setToolTip(u'Recarga la lista de programas')
        action_reload.triggered.connect(self.refresh_episodes)
        menu_appl.addAction(action_reload)

        icon = self.style().standardIcon(QStyle.SP_FileDialogDetailedView)
        action_preferences = QAction(icon, u'&Preferencias', self)
        action_preferences.triggered.connect(self.open_preferences)
        action_preferences.setToolTip(
            u'Configurar distintos parámetros del programa')
        menu_appl.addAction(action_preferences)

        menu_appl.addSeparator()

        icon = self.style().standardIcon(QStyle.SP_MessageBoxInformation)
        _act = QAction(icon, '&Acerca de', self)
        _act.triggered.connect(self.open_about_dialog)
        _act.setToolTip(u'Muestra información de la aplicación')
        menu_appl.addAction(_act)

        icon = self.style().standardIcon(QStyle.SP_DialogCloseButton)
        _act = QAction(icon, '&Salir', self)
        _act.setShortcut('Ctrl+Q')
        _act.setToolTip(u'Sale de la aplicación')
        _act.triggered.connect(self.on_close)
        menu_appl.addAction(_act)

        # program menu
        menu_prog = menubar.addMenu(u'&Programa')

        icon = self.style().standardIcon(QStyle.SP_ArrowDown)
        self.action_download = QAction(icon, '&Descargar', self)
        self.action_download.setShortcut('Ctrl+D')
        self.action_download.setEnabled(False)
        self.action_download.setToolTip(TTIP_DOWNLOAD_D)
        self.action_download.triggered.connect(self.download_episode)
        menu_prog.addAction(self.action_download)

        icon = self.style().standardIcon(QStyle.SP_MediaPlay)
        self.action_play = QAction(icon, '&Reproducir', self)
        self.action_play.setEnabled(False)
        self.action_play.setToolTip(TTIP_PLAY_D)
        self.action_play.triggered.connect(self.on_play_action)
        menu_prog.addAction(self.action_play)

        # toolbar for buttons
        toolbar = self.addToolBar('main')
        toolbar.addAction(self.action_download)
        toolbar.addAction(self.action_play)
        toolbar.addSeparator()
        toolbar.addAction(action_reload)
        toolbar.addAction(action_preferences)

        # filter text and button, to the right
        spacer = QWidget()
        spacer.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        toolbar.addWidget(spacer)
        toolbar.addWidget(QLabel(u"Filtro: "))
        self.filter_line = QLineEdit()
        self.filter_line.setMaximumWidth(150)
        self.filter_line.textChanged.connect(self.on_filter_changed)
        toolbar.addWidget(self.filter_line)
        self.filter_cbox = QCheckBox(u"Sólo descargados")
        self.filter_cbox.stateChanged.connect(self.on_filter_changed)
        toolbar.addWidget(self.filter_cbox)
        QShortcut(QKeySequence("Ctrl+F"), self, self.filter_line.setFocus)

        # if needed, a warning that stuff needs to be configured
        icon = self.style().standardIcon(QStyle.SP_MessageBoxWarning)
        m = u"Necesita configurar algo; haga click aquí para abrir el wizard"
        self.needsomething_alert = QAction(icon, m, self)
        self.needsomething_alert.triggered.connect(self._start_wizard)
        toolbar.addAction(self.needsomething_alert)
        if not config.get('nowizard'):
            self._start_wizard()
        self._review_need_something_indicator()

    def _start_wizard(self, _=None):
        """Start the wizard if needed."""
        if not self.have_config() or not self.have_metadata():
            dlg = wizard.WizardDialog(self)
            dlg.exec_()
        self._review_need_something_indicator()

    def on_filter_changed(self, _):
        """The filter text has changed, apply it in the episodes list."""
        text = self.filter_line.text()
        cbox = self.filter_cbox.checkState()
        self.episodes_list.set_filter(text, cbox)

        # after applying filter, nothing is selected, so check buttons
        # (easiest way to clean them all)
        self.check_download_play_buttons()

    def _review_need_something_indicator(self):
        """Hide/show/enable/disable different indicators if need sth."""
        needsomething = bool(not self.have_config() or
                             not self.have_metadata())
        self.needsomething_alert.setVisible(needsomething)

    def shutdown(self):
        """Stop everything and quit.

        This shutdown con be called at any time, even on init, so we have
        extra precautions about which attributes we have.
        """
        signal.emit('save_state')
        config.save()
        self.finished = True

        programs_data = getattr(self, 'programs_data', None)
        if programs_data is not None:
            programs_data.save()

        downloaders = getattr(self, 'downloaders', {})
        for downloader in downloaders.itervalues():
            downloader.shutdown()

        # bye bye
        self.app_quit()

    def on_close(self, _):
        """Close signal."""
        if self._should_close():
            self.shutdown()

    def closeEvent(self, event):
        """All is being closed."""
        if self._should_close():
            self.shutdown()
        else:
            event.ignore()

    def _should_close(self):
        """Still time to decide if want to close or not."""
        logger.info("Attempt to close the program")
        pending = self.episodes_download.pending()
        if not pending:
            # all fine, save all and quit
            logger.info("Saving states and quitting")
            return True
        logger.debug("Still %d active downloads when trying to quit", pending)

        # stuff pending
        m = (u"Hay programas todavía en proceso de descarga!\n"
             u"¿Seguro quiere salir del programa?")
        QMB = QMessageBox
        dlg = QMB(u"Guarda!", m, QMB.Question, QMB.Yes, QMB.No, QMB.NoButton)
        opt = dlg.exec_()
        if opt != QMB.Yes:
            logger.info("Quit cancelled")
            return False

        # quit anyway, put all downloading and pending episodes to none
        logger.info("Fixing episodes, saving state and exiting")
        for program in self.programs_data.values():
            state = program.state
            if state == Status.waiting or state == Status.downloading:
                program.state = Status.none
        return True

    def show_message(self, err_type, text):
        """Show different messages to the user."""
        if self.finished:
            logger.debug("Ignoring message: %r", text)
            return
        logger.debug("Showing a message: %r", text)

        # error text can be produced by windows, try to to sanitize it
        if isinstance(text, str):
            try:
                text = text.decode("utf8")
            except UnicodeDecodeError:
                try:
                    text = text.decode("latin1")
                except UnicodeDecodeError:
                    text = repr(text)

        QMB = QMessageBox
        dlg = QMB(u"Atención: " + err_type, text, QMB.Warning,
                  QMB.Ok, QMB.NoButton, QMB.NoButton)
        dlg.exec_()

    def refresh_episodes(self, _=None):
        """Update and refresh episodes."""
        ue = update.UpdateEpisodes(self)
        ue.interactive()

    def download_episode(self, _=None):
        """Download the episode(s)."""
        episode_ids = self.episodes_list.selected_items()
        for episode_id in episode_ids:
            episode = self.programs_data[episode_id]
            self.queue_download(episode)

    @defer.inline_callbacks
    def queue_download(self, episode):
        """User indicated to download something."""
        logger.debug("Download requested of %s", episode)
        if episode.state != Status.none:
            logger.debug("Download denied, episode %s is not in downloadeable "
                         "state.", episode.episode_id)
            return

        # queue
        self.episodes_download.append(episode)
        self.episodes_list.episode_info.update(episode)
        self.check_download_play_buttons()
        if self.episodes_download.downloading:
            return

        logger.debug("Downloads: starting")
        while self.episodes_download.pending():
            episode = self.episodes_download.prepare()
            try:
                filename, episode = yield self._episode_download(episode)
            except CancelledError:
                logger.debug("Got a CancelledError!")
                self.episodes_download.end(error=u"Cancelado")
            except BadCredentialsError:
                logger.debug("Bad credentials error!")
                msg = (u"Error con las credenciales: hay que configurar "
                       u"usuario y clave correctos")
                self.show_message('BadCredentialsError', msg)
                self.episodes_download.end(error=msg)
            except EncuentroError, e:
                orig_exc = e.orig_exc
                msg = "%s(%s)" % (orig_exc, e)
                err_type = e.__class__.__name__
                logger.exception("Custom Encuentro error: %s (%r)",
                                 e, orig_exc)
                notify(err_type, msg)
                self.episodes_download.end(error=u"Error: " + msg)
            except Exception, e:
                logger.exception("Unknown download error: %s (%r)", e, e)
                err_type = e.__class__.__name__
                notify(err_type, str(e))
                self.episodes_download.end(error=u"Error: " + str(e))
            else:
                logger.debug("Episode downloaded: %s", episode)
                self.episodes_download.end()
                episode.filename = filename

            # check buttons
            self.check_download_play_buttons()

            # adjust the episode info only if it's still showing this one
            self.episodes_list.episode_info.update(episode, force_change=False)

        logger.debug("Downloads: finished")

    @defer.inline_callbacks
    def _episode_download(self, episode):
        """Effectively download an episode."""
        logger.debug("Effectively downloading episode %s", episode.episode_id)
        self.episodes_download.start()

        # download!
        downloader = self.downloaders[episode.downtype]
        season = getattr(episode, 'season', None)  # wasn't always there
        fname = yield downloader.download(
            episode.channel, episode.section, season, episode.title,
            episode.url, self.episodes_download.progress)
        episode_name = u"%s - %s - %s" % (episode.channel, episode.section,
                                          episode.composed_title)
        notify(u"Descarga finalizada", episode_name)
        defer.return_value((fname, episode))

    def open_preferences(self, _=None):
        """Open the preferences dialog."""
        dlg = preferences.PreferencesDialog()
        dlg.exec_()
        # after dialog closes, config changed, so review indicators
        self._review_need_something_indicator()
        safecfg = config.sanitized_config()
        logger.debug("Configuration changed: %s", safecfg)

    def check_download_play_buttons(self):
        """Set both buttons state according to the selected episodes."""
        episode_ids = self.episodes_list.selected_items()

        # 'play' button should be enabled if only one row is selected and
        # its state is 'downloaded'
        play_enabled = False
        if len(episode_ids) == 1:
            episode = self.programs_data[episode_ids[0]]
            if episode.state == Status.downloaded:
                play_enabled = True
        self.action_play.setEnabled(play_enabled)
        ttip = TTIP_PLAY_E if play_enabled else TTIP_PLAY_D
        self.action_play.setToolTip(ttip)

        # 'download' button should be enabled if at least one of the selected
        # rows is in 'none' state, and if config is ok
        download_enabled = False
        if self.have_config():
            for episode_id in episode_ids:
                episode = self.programs_data[episode_id]
                if episode.state == Status.none:
                    download_enabled = True
                    break
        ttip = TTIP_DOWNLOAD_E if download_enabled else TTIP_DOWNLOAD_D
        self.action_download.setEnabled(download_enabled)
        self.action_download.setToolTip(ttip)

    def on_play_action(self, _=None):
        """Play the selected episode."""
        episode_ids = self.episodes_list.selected_items()
        if len(episode_ids) != 1:
            raise ValueError("Wrong call to play_episode, with %d selections"
                             % len(episode_ids))
        episode_id = episode_ids[0]
        episode = self.programs_data[episode_id]
        self.play_episode(episode)

    def play_episode(self, episode):
        """Play an episode."""
        downloaddir = config.get('downloaddir', '')
        filename = os.path.join(downloaddir, episode.filename)

        logger.info("Play requested of %s", episode)
        if os.path.exists(filename):
            # pass file:// url with absolute path
            fullpath = 'file://' + os.path.abspath(filename)
            logger.info("Playing %r", fullpath)
            multiplatform.open_file(fullpath)
        else:
            logger.warning("Aborted playing, file not found: %r", filename)
            msg = (u"No se encontró el archivo para reproducir: " +
                   repr(filename))
            self.show_message('Error al reproducir', msg)
            episode.state = Status.none
            self.episodes_list.set_color(episode)

    def cancel_download(self, episode):
        """Cancel the downloading of an episode."""
        logger.info("Cancelling download of %s", episode)
        self.episodes_download.cancel()
        downloader = self.downloaders[episode.downtype]
        downloader.cancel()

    def unqueue_download(self, episode):
        """Remove the episode from the download queue."""
        logger.info("Unqueueing %s", episode)
        self.episodes_download.unqueue(episode)

    def open_about_dialog(self):
        """Show the about dialog."""
        version = self.version if self.version else u"(?)"
        title = "Encuentro v" + version
        text = ABOUT_TEXT % (version,)
        QMessageBox.about(self, title, text)



# Copyright 2013-2014 Facundo Batista
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# For further info, check  https://launchpad.net/encuentro

"""Main entry point, and initialization of everything we can."""

import logging
import os
import sys

from encuentro import multiplatform
from encuentro.config import config
from encuentro.ui.main import MainUI

# we put here EpisodeData only for legacy reasons: unpickle of old pickles
# will try to load EpisodeData from this namespace
from encuentro.data import EpisodeData  # NOQA

logger = logging.getLogger('encuentro.init')

from PyQt4.QtGui import QApplication, QIcon


def start(version):
    """Rock and roll."""
    # set up config
    fname = os.path.join(multiplatform.config_dir, 'encuentro.conf')
    print "Using configuration file:", repr(fname)
    logger.info("Using configuration file: %r", fname)
    config.init(fname)

    # the order of the lines hereafter are very precise, don't mess with them
    app = QApplication(sys.argv)
    icon = QIcon(multiplatform.get_path("encuentro/logos/icon-192.png"))
    app.setWindowIcon(icon)

    MainUI(version, app.quit)
    sys.exit(app.exec_())



# Copyright 2011-2013 Facundo Batista
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# For further info, check  https://launchpad.net/encuentro

"""Logging set up."""

import logging
import os
import sys
import traceback

from logging.handlers import RotatingFileHandler

import xdg.BaseDirectory


class CustomRotatingFH(RotatingFileHandler):
    """Rotating handler that starts a new file for every run."""

    def __init__(self, *args, **kwargs):
        RotatingFileHandler.__init__(self, *args, **kwargs)
        self.doRollover()


def exception_handler(exc_type, exc_value, tb):
    """Handle an unhandled exception."""
    exception = traceback.format_exception(exc_type, exc_value, tb)
    msg = "".join(exception)
    print >> sys.stderr, msg

    # log
    logger = logging.getLogger('encuentro')
    logger.error("Unhandled exception!\n%s", msg)


def get_filename():
    """Return the log file name."""
    return os.path.join(xdg.BaseDirectory.xdg_cache_home,
                        'encuentro', 'encuentro.log')


def set_up(verbose):
    """Set up the logging."""

    logfile = get_filename()
    print "Saving logs to", repr(logfile)
    logfolder = os.path.dirname(logfile)
    if not os.path.exists(logfolder):
        os.makedirs(logfolder)

    logger = logging.getLogger('encuentro')
    handler = CustomRotatingFH(logfile, maxBytes=1e6, backupCount=10)
    logger.addHandler(handler)
    formatter = logging.Formatter("%(asctime)s  %(name)-22s  "
                                  "%(levelname)-8s %(message)s")
    handler.setFormatter(formatter)
    logger.setLevel(logging.DEBUG)

    if verbose:
        handler = logging.StreamHandler()
        handler.setFormatter(formatter)
        logger.addHandler(handler)

    # hook the exception handler
    sys.excepthook = exception_handler



# -*- coding: utf8 -*-
# Copyright 2013 Facundo Batista
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# For further info, check  https://launchpad.net/encuentro

"""The wizard that guides the user for the initial setup."""

import logging

from PyQt4.QtGui import (
    QCheckBox,
    QDialog,
    QDialogButtonBox,
    QLabel,
    QPushButton,
    QVBoxLayout,
)

from encuentro.config import config

logger = logging.getLogger('encuentro.wizard')

TEXT_INIT = u"""
Bienvenido al visor de contenido del Canal Encuentro y otros.
Para poder usar el programa debe primero configurarlo!
"""

TEXT_EPISODES = u"""
Primero tiene que actualizar la lista de episodios:
puede actualizar la lista ahora desde esta misma ventana
y en cualquier momento desde el menú del programa.
"""

TEXT_CONFIG = u"""
Para poder descargar los programas de los distintos backends tiene que
configurar algunos con usuario y clave; puede configurar el sistema
ahora desde esta misma ventana o en cualquier momento desde el menú
del programa.
"""

TEXT_HAPPY_END = u"""
Felicitaciones, el programa está listo para usar :)
"""

TEXT_SAD_END = u"""
¡Ya podes usar el programa!
(aunque te falta actualizar y/o configurar algo)
"""

# The steps for the wizard
# - the text to show to the user
# - the method to decide if the step should be ignored ("self._ign_"...)
# - the text for the action button
# - the action when clicked (will compose method with "self._act_")
STEPS = [
    (TEXT_INIT, None, None, None),
    (TEXT_EPISODES, "episode", u"Actualizar", "update"),
    (TEXT_CONFIG, "config", u"Configurar", "configure"),
    (None, None, None, None),
]


class WizardDialog(QDialog):
    """The dialog for update."""
    def __init__(self, main_window):
        super(WizardDialog, self).__init__()
        self.mw = main_window
        vbox = QVBoxLayout(self)

        # label and checkbox
        self.main_text = QLabel(u"init text")
        vbox.addWidget(self.main_text)
        self.notthisagain = QCheckBox(u"No mostrar automáticamente esta ayuda")
        nowizard = config.get('nowizard', False)
        self.notthisagain.setCheckState(nowizard)
        self.notthisagain.stateChanged.connect(self._notthisagain_toggled)
        vbox.addWidget(self.notthisagain)

        # buttons
        bbox = QDialogButtonBox()
        self.navbut_actn = QPushButton(u"init text")
        bbox.addButton(self.navbut_actn, QDialogButtonBox.ActionRole)
        self.navbut_prev = QPushButton(u"Anterior")
        bbox.addButton(self.navbut_prev, QDialogButtonBox.ActionRole)
        self.navbut_next = QPushButton(u"Siguiente")
        bbox.addButton(self.navbut_next, QDialogButtonBox.ActionRole)
        vbox.addWidget(bbox)

        self.show()
        self.step = 0
        self._move(0)

    def _notthisagain_toggled(self, state):
        """The "not this again" checkbutton togled state."""
        logger.info("Configuring 'nowizard' to %s", state)
        config['nowizard'] = state

    def _move(self, delta_step):
        """The engine for the wizard steps."""
        self.step += delta_step
        logger.debug("Entering into step %d", self.step)
        (text, ign_func, act_label, act_func) = STEPS[self.step]
        # if this step should be ignored, just leave
        if ign_func is not None:
            m = getattr(self, "_ign_" + ign_func)
            if m():
                # keep going
                return self._move(delta_step)

        # adjust navigation buttons
        if self.step == 0:
            self.navbut_prev.setEnabled(False)
            self.navbut_next.setText(u"Siguiente")
            self.navbut_next.clicked.disconnect()
            self.navbut_next.clicked.connect(lambda: self._move(1))
            self.notthisagain.show()
        elif self.step == len(STEPS) - 1:
            self.navbut_prev.setEnabled(True)
            self.navbut_prev.clicked.disconnect()
            self.navbut_prev.clicked.connect(lambda: self._move(-1))
            self.navbut_next.setText(u"Terminar")
            self.navbut_next.clicked.disconnect()
            self.navbut_next.clicked.connect(self.accept)
            self.notthisagain.hide()
        else:
            self.navbut_prev.setEnabled(True)
            self.navbut_prev.clicked.disconnect()
            self.navbut_prev.clicked.connect(lambda: self._move(-1))
            self.navbut_next.setText(u"Siguiente")
            self.navbut_next.clicked.disconnect()
            self.navbut_next.clicked.connect(lambda: self._move(1))
            self.notthisagain.hide()

        # adjust main text and action button
        if self.step == len(STEPS) - 1:
            if self.mw.have_metadata() and self.mw.have_config():
                self.main_text.setText(TEXT_HAPPY_END)
            else:
                self.main_text.setText(TEXT_SAD_END)
        else:
            self.main_text.setText(text)

        if act_label is None:
            self.navbut_actn.hide()
        else:
            self.navbut_actn.show()
            self.navbut_actn.setText(act_label)
            method_to_call = getattr(self, "_act_" + act_func)
            self.navbut_actn.clicked.disconnect()
            self.navbut_actn.clicked.connect(method_to_call)

    def _act_configure(self, _):
        """Open the config dialog."""
        self.mw.open_preferences()

    def _act_update(self, *a):
        """Open the update dialog."""
        self.mw.refresh_episodes()

    def _ign_episode(self):
        """Tell if the episode step should be ignored."""
        return self.mw.have_metadata()

    def _ign_config(self):
        """Tell if the configure step should be ignored."""
        return self.mw.have_config()


if __name__ == '__main__':
    import sys

    from PyQt4.QtGui import QApplication
    app = QApplication(sys.argv)
    app.have_metadata = lambda: False
    app.have_config = lambda: False

    frame = WizardDialog(app)
    frame.show()
    frame.exec_()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#
# Copyright 2011 Facundo Batista
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# For further info, check  https://launchpad.net/encuentro

"""The package."""

import sys

# special import before any other imports to configure GUI to use API 2
import sip
for name in "QDate QDateTime QString QTextStream QTime QUrl QVariant".split():
    sip.setapi(name, 2)   # API v2 FTW!


IMPORT_MSG = u"""
ERROR! Problema al importar %(module)r

Probablemente falte instalar una dependencia.  Se necesita tener instalado
el paquete %(package)r versión %(version)s o superior.
"""


class NiceImporter(object):
    """Show nicely successful and errored imports."""
    def __init__(self, module, package, version):
        self.module = module
        self.package = package
        self.version = version

    def __enter__(self):
        pass

    def _get_version(self):
        """Get the version of a module."""
        mod = sys.modules[self.module]
        for attr in ('version', '__version__', 'ver', 'PYQT_VERSION_STR'):
            v = getattr(mod, attr, None)
            if v is not None:
                return v
        return "<desconocida>"

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is None:
            version = self._get_version()
            print "Modulo %r importado ok, version %r" % (self.module, version)
        else:
            print IMPORT_MSG % dict(module=self.module, package=self.package,
                                    version=self.version)

        # consume the exception!
        return True


# test the packages
with NiceImporter('xdg', 'python-xdg', '0.15'):
    import xdg  # NOQA
with NiceImporter('requests', 'python-requests', '2.2.1'):
    import requests  # NOQA
with NiceImporter('PyQt4.QtCore', 'PyQt4', '4.9.1'):
    import PyQt4.QtCore  # NOQA
with NiceImporter('defer', 'python-defer', '1.0.6'):
    import defer  # NOQA



# -*- coding: utf8 -*-

# Copyright 2014 Facundo Batista
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# For further info, check  https://launchpad.net/encuentro

"""The notification-to-the-desktop subsystem."""

from encuentro.config import config


_ERRMSG = u"""
ERROR! Problema al importar 'pynotify' - No es "estrictamente necesario, pero
si lo instala tendrá algunas notificaciones en el escritorio.
"""


class _Notifier(object):
    """A notifier that defers the import as much as possible.

    This is because importing 'pynotify' while PyQt is still starting causes
    everything to segfault.
    """
    def __init__(self):
        self._inited = False

    def _init(self):
        """Initialize everything."""
        self._inited = True
        try:
            import pynotify
        except ImportError:
            print _ERRMSG
            self._notify = lambda t, m: None
        else:
            pynotify.init("Encuentro")

            def _f(title, message):
                """The method that will really notify."""
                if config.get('notification', True):
                    n = pynotify.Notification(title, message)
                    n.show()

            self._notify = _f

    def __call__(self, title, message):
        if not self._inited:
            self._init()
        self._notify(title, message)


notify = _Notifier()



# -*- coding: UTF-8 -*-

# Copyright 2013 Facundo Batista
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# For further info, check  https://launchpad.net/encuentro

"""The system configuration."""

import logging
import os
import pickle

from encuentro import utils


logger = logging.getLogger('encuentro.config')

# these are the configuration variables/parameters that on one hand should
# be stored in a keyring (if available), and on the other hand must not be
# logged by the system
SECURITY_CONFIG = ['user', 'password']


class _Config(dict):
    """The configuration."""

    SYSTEM = 'system'

    def __init__(self):
        self._fname = None
        super(_Config, self).__init__()

    def sanitized_config(self):
        """Return a copied config, sanitized to log."""
        safecfg = self.copy()
        for secure in SECURITY_CONFIG:
            if secure in safecfg:
                safecfg[secure] = '<hidden>'
        return safecfg

    def init(self, fname):
        """Initialize and load config."""
        self._fname = fname
        if not os.path.exists(fname):
            # default to an almost empty dict
            self[self.SYSTEM] = {}
            logger.debug("File not found, starting empty")
            return

        with open(fname, 'rb') as fh:
            saved_dict = pickle.load(fh)
            logger.debug("Loaded: %s", self.sanitized_config())
        self.update(saved_dict)

        # for compatibility, put the system container if not there
        if self.SYSTEM not in self:
            self[self.SYSTEM] = {}

    def save(self):
        """Save the config to disk."""
        # we don't want to pickle this class, but the dict itself
        raw_dict = self.copy()
        logger.debug("Saving: %s", self.sanitized_config())
        with utils.SafeSaver(self._fname) as fh:
            pickle.dump(raw_dict, fh)


class _Signal(object):
    """Custom signals.

    Decorate a function to be called when signal is emitted.
    """

    def __init__(self):
        self.store = {}

    def register(self, method):
        """Register a method."""
        self.store.setdefault(method.__name__, []).append(method)

    def emit(self, name):
        """Call the registered methods."""
        meths = self.store.get(name, [])
        for meth in meths:
            meth()


config = _Config()
signal = _Signal()



# Copyright 2013-2014 Facundo Batista
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# For further info, check  https://launchpad.net/encuentro

"""Classes to interface to and persist the episodes data."""

import cgi
import logging
import os
import pickle

from base64 import b64decode
from unicodedata import normalize

from encuentro import utils
from encuentro.ui import dialogs

logger = logging.getLogger('encuentro.data')


class Status(object):
    """Status constants."""
    none = 'none'
    waiting = 'waiting'
    downloading = 'downloading'
    downloaded = 'downloaded'


_normalize_cache = {}


def _search_normalizer(char):
    """Normalize always to one char length."""
    try:
        return _normalize_cache[char]
    except KeyError:
        norm = normalize('NFKD', char).encode('ASCII', 'ignore').lower()
        if not norm:
            norm = '?'
        _normalize_cache[char] = norm
        return norm


def prepare_to_filter(text):
    """Prepare a text to filter.

    It receives unicode, but return simple lowercase ascii.
    """
    return ''.join(_search_normalizer(c) for c in text)


class EpisodeData(object):
    """Episode data."""

    # these is for the attributes to be here when unpickling old instances
    image_url = None
    downtype = None
    image_data = None
    subtitle = None

    def __init__(self, channel, section, title, duration, description,
                 episode_id, url, image_url, state=None, progress=None,
                 filename=None, downtype=None, season=None,
                 image_data=None, subtitle=None):
        self.channel = channel
        self.section = section
        self.season = None if season is None else cgi.escape(season)
        self.title = cgi.escape(title)
        self.duration = duration
        self.description = description
        self.subtitle = subtitle
        self.episode_id = episode_id

        # build a nice string to show in the GUI
        if self.season:
            self.composed_title = self.season + u": " + self.title
        else:
            self.composed_title = self.title

        # urls are bytes!
        self.url = str(url)
        self.image_url = str(image_url)

        # image data is encoded in base64
        self.image_data = None if image_data is None else b64decode(image_data)

        self.state = Status.none if state is None else state
        self.progress = progress
        self.filename = filename
        self.to_filter = None
        self.downtype = downtype

        # cache the processed title
        self._normalized_title = prepare_to_filter(self.composed_title)

    @property
    def normalized_title(self):
        """Get the normalized title, if already have it, or calculate it.

        This attribute may not be present because old pickled instances didn't
        have it.
        """
        if not hasattr(self, '_normalized_title'):
            self._normalized_title = prepare_to_filter(self.composed_title)
        return self._normalized_title

    def update(self, channel, section, title, duration, description,
               episode_id, url, image_url, state=None, progress=None,
               filename=None, downtype=None, season=None,
               image_data=None, subtitle=None):
        """Update the episode data."""
        self.channel = channel
        self.section = section
        self.season = None if season is None else cgi.escape(season)
        self.title = cgi.escape(title)
        self.duration = duration
        self.description = description
        self.subtitle = subtitle
        self.episode_id = episode_id

        # build a nice string to show in the GUI
        if self.season:
            self.composed_title = self.season + u": " + self.title
        else:
            self.composed_title = self.title

        # urls are bytes!
        self.url = str(url)
        self.image_url = str(image_url)

        # image data is encoded in base64
        self.image_data = None if image_data is None else b64decode(image_data)

        self.state = Status.none if state is None else state
        self.progress = progress
        self.filename = filename
        self.downtype = downtype

        # cache the processed title, overwritting what may be old from the past
        self._normalized_title = prepare_to_filter(self.composed_title)

    def filter_params(self, text, only_downloaded):
        """Return the filtering params.

        If should filter, it will return (pos1, pos2) (both in None if it only
        filters by only_downloaded). If it should not filter, will return None.
        """
        if only_downloaded and self.state != Status.downloaded:
            # need downloaded ones, sorry
            return

        t = self.normalized_title
        pos1 = t.find(text)
        if pos1 == -1:
            # need to match text, sorry
            return

        # return boundaries
        pos2 = pos1 + len(text)
        return (pos1, pos2)

    def __str__(self):
        args = (self.episode_id, self.state, self.channel,
                self.section, self.title)
        return "<EpisodeData [%s] (%s) %r (%r): %r>" % args


class ProgramsData(object):
    """Holder / interface for programs data."""

    # more recent version of the in-disk data
    last_programs_version = 2

    def __init__(self, main_window, filename):
        self.main_window = main_window
        self.filename = filename
        print "Using data file:", repr(filename)
        logger.info("Using data file: %r", filename)

        self.version = None
        self.data = None
        self.reset_config_from_migration = False
        self.load()
        self.migrate()
        logger.info("Episodes metadata loaded (total %d)", len(self.data))

    def merge(self, new_data, episodes_widget):
        """Merge new data to current programs data."""
        for d in new_data:
            names = ['channel', 'section', 'title', 'duration',
                     'description', 'episode_id', 'url', 'image_url',
                     'downtype', 'season', 'image_data', 'subtitle']
            values = dict((name, d.get(name)) for name in names)
            episode_id = d['episode_id']

            try:
                ed = self.data[episode_id]
            except KeyError:
                self.data[episode_id] = EpisodeData(**values)
            else:
                ed.update(**values)

        episodes_widget.reload_episodes()
        self.save()

    def load(self):
        """Load the data from the file."""
        # if not file, all empty
        if not os.path.exists(self.filename):
            self.data = {}
            self.version = self.last_programs_version
            return

        # get from the file
        with open(self.filename, 'rb') as fh:
            try:
                loaded_programs_data = pickle.load(fh)
            except Exception, err:
                logger.warning("ERROR while opening the pickled data: %s", err)
                self.data = {}
                self.version = self.last_programs_version
                return

        # check pre-versioned data
        if isinstance(loaded_programs_data, dict):
            # pre-versioned data
            self.version = 0
            self.data = loaded_programs_data
        else:
            self.version, self.data = loaded_programs_data

    def migrate(self):
        """Migrate metadata if needed."""
        if self.version == self.last_programs_version:
            logger.info("Metadata is updated, nothing to migrate")
            return

        if self.version > self.last_programs_version:
            raise ValueError("Data is newer than code! %s" % (self.version,))

        # migrate
        if self.version == 0:
            logger.info("Migrating from version 0")
            # actually, from 0, no migration is possible, we
            # need to tell the user the ugly truth
            dlg = dialogs.ForceUpgradeDialog()
            should_quit = dlg.exec_()
            if should_quit:
                self.main_window.shutdown()
                return
            # if user accessed to go on, don't really need to migrate
            # anything, as *all* the code is to support the new metadata
            # version only, so just remove it and mark the usr/pass config
            # to be removed
            self.version = self.last_programs_version
            self.reset_config_from_migration = True
            self.data = {}
            return

        if self.version == 1:
            logger.info("Migrating from version 1")
            self.version = self.last_programs_version
            for epis_id, episode in self.data.items():
                episode.composed_title = episode.title
            return

        raise ValueError("Don't know how to migrate from %r" % (self.version,))

    def __str__(self):
        return "<ProgramsData ver=%r len=%d>" % (self.version, len(self.data))

    def __nonzero__(self):
        return bool(self.data)

    def __getitem__(self, pos):
        return self.data[pos]

    def __setitem__(self, pos, value):
        self.data[pos] = value

    def values(self):
        """Return the iter values of the data."""
        return self.data.itervalues()

    def __len__(self):
        """The length."""
        return len(self.data)

    def items(self):
        """Return the iter items of the data."""
        return self.data.iteritems()

    def save(self):
        """Save to disk."""
        to_save = (self.last_programs_version, self.data)
        with utils.SafeSaver(self.filename) as fh:
            pickle.dump(to_save, fh)



# Copyright 2013 Facundo Batista
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# For further info, check  https://launchpad.net/encuentro

"""Some useful functions."""

import defer
import os

from PyQt4 import QtNetwork, QtCore

_qt_network_manager = QtNetwork.QNetworkAccessManager()


class _Downloader(object):
    """An asynch downloader that fires a deferred with data when done."""
    def __init__(self, url):
        self.deferred = defer.Deferred()
        self.deferred._store_it_because_qt_needs_or_wont_work = self
        request = QtNetwork.QNetworkRequest(QtCore.QUrl(url))

        self.req = _qt_network_manager.get(request)
        self.req.error.connect(self.deferred.errback)
        self.req.finished.connect(self.end)

    def end(self):
        """Send data through the deferred, if wasn't fired before."""
        img_data = self.req.read(self.req.bytesAvailable())
        content_type = self.req.header(
            QtNetwork.QNetworkRequest.ContentTypeHeader)
        data = (content_type, img_data)
        if not self.deferred.called:
            self.deferred.callback(data)


def download(url):
    """Deferredly download an URL, non blocking."""
    d = _Downloader(url)
    return d.deferred


class SafeSaver(object):
    """A safe saver to disk.

    It saves to a .tmp and moves into final destination, and other
    considerations.
    """

    def __init__(self, fname):
        self.fname = fname
        self.tmp = fname + ".tmp"
        self.fh = None

    def __enter__(self):
        self.fh = open(self.tmp, 'wb')
        return self.fh

    def __exit__(self, *exc_data):
        self.fh.close()

        # only move into final destination if all went ok
        if exc_data == (None, None, None):
            if os.path.exists(self.fname):
                # in Windows we need to remove the old file first
                os.remove(self.fname)
            os.rename(self.tmp, self.fname)


if __name__ == "__main__":
    import sys
    app = QtCore.QCoreApplication(sys.argv)
    _url = "http://www.taniquetil.com.ar/facundo/imgs/felu-camagrande.jpg"

    @defer.inline_callbacks
    def _download():
        """Download."""
        deferred = download(_url)
        data = yield deferred
        print "All done!", len(data), type(data)
    _download()
    sys.exit(app.exec_())



# -*- coding: UTF-8 -*-

# Copyright 2013-2014 Facundo Batista
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# For further info, check  https://launchpad.net/encuentro

"""A throbber."""

from PyQt4.QtGui import (
    QLabel,
    QMovie,
)
from PyQt4.QtCore import Qt

from encuentro import multiplatform


class Throbber(QLabel):
    """A throbber."""
    def __init__(self):
        super(Throbber, self).__init__()
        self.setAlignment(Qt.AlignCenter)
        fname = multiplatform.get_path("encuentro/ui/media/throbber.gif")
        self._movie = QMovie(fname)
        self.setMovie(self._movie)

    def hide(self):
        """Overload to control the movie."""
        self._movie.stop()
        super(Throbber, self).hide()

    def show(self):
        """Overload to control the movie."""
        self._movie.start()
        super(Throbber, self).show()




# Copyright 2011-2013 Facundo Batista
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# For further info, check  https://launchpad.net/encuentro

"""Configuration file for Apport."""

from apport.hookutils import attach_related_packages, attach_file_if_exists

from encuentro import logger


def add_info(report):
    """Add info to the report."""
    # attach the log
    fname = logger.get_filename()
    attach_file_if_exists(report, fname, "EncuentroLog")

    # info about dependencies
    packages = ["python-requests", "python-defer", "python-qt4", "python-xdg"]
    attach_related_packages(report, packages)



# -*- coding: utf8 -*-

# Copyright 2013 Facundo Batista
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# For further info, check  https://launchpad.net/encuentro

"""The preferences dialog."""

import os
import sys
import logging

from PyQt4.QtGui import (
    QCheckBox,
    QCompleter,
    QDialog,
    QDialogButtonBox,
    QDirModel,
    QFileDialog,
    QGridLayout,
    QLabel,
    QLineEdit,
    QPushButton,
    QTabWidget,
    QVBoxLayout,
    QWidget,
)
from PyQt4.QtCore import Qt, QDir

from encuentro.config import config

logger = logging.getLogger('encuentro.preferences')

URL_CONECTATE = (
    "http://registro.educ.ar/cuentas/registro/index?servicio=conectate"
)


class GeneralPreferences(QWidget):
    """The general preferences input."""
    def __init__(self):
        super(GeneralPreferences, self).__init__()
        grid = QGridLayout(self)
        grid.setSpacing(20)
        grid.setColumnStretch(1, 10)

        # directory auto completer
        completer = QCompleter(self)
        dirs = QDirModel(self)
        dirs.setFilter(QDir.AllDirs | QDir.NoDotAndDotDot)
        completer.setModel(dirs)
        completer.setCaseSensitivity(Qt.CaseInsensitive)
        completer.setCompletionMode(QCompleter.PopupCompletion)

        l = QLabel(
            u"<b>Ingresá el directorio donde descargar los videos...</b>")
        l.setTextFormat(Qt.RichText)
        grid.addWidget(l, 0, 0, 1, 2)

        grid.addWidget(QLabel(u"Descargar en:"), 1, 0, 2, 1)
        prv = config.get('downloaddir', '')
        self.downloaddir_entry = QLineEdit(prv)
        self.downloaddir_entry.setCompleter(completer)
        self.downloaddir_entry.setPlaceholderText(u'Ingresá un directorio')
        grid.addWidget(self.downloaddir_entry, 1, 1, 2, 2)

        self.downloaddir_buttn = QPushButton(u"Elegir un directorio")
        self.downloaddir_buttn.clicked.connect(self._choose_dir)
        grid.addWidget(self.downloaddir_buttn, 2, 1, 3, 2)

        self.autoreload_checkbox = QCheckBox(
            u"Recargar automáticamente la lista de episodios al iniciar")
        prv = config.get('autorefresh', False)
        self.autoreload_checkbox.setChecked(prv)
        grid.addWidget(self.autoreload_checkbox, 3, 0, 4, 2)

        self.shownotifs_checkbox = QCheckBox(
            u"Mostrar una notificación cuando termina cada descarga")
        prv = config.get('notification', True)
        self.shownotifs_checkbox.setChecked(prv)
        grid.addWidget(self.shownotifs_checkbox, 4, 0, 5, 2)

    def _choose_dir(self):
        """Choose a directory using a dialog."""
        resp = QFileDialog.getExistingDirectory(self, '',
                                                os.path.expanduser("~"))
        if resp:
            self.downloaddir_entry.setText(resp)

    def get_config(self):
        """Return the config for this tab."""
        d = {}
        d['downloaddir'] = self.downloaddir_entry.text()
        d['autorefresh'] = self.autoreload_checkbox.isChecked()
        d['notification'] = self.shownotifs_checkbox.isChecked()
        return d


class ConectatePreferences(QWidget):
    """The preferences for Conectate backend."""
    def __init__(self):
        super(ConectatePreferences, self).__init__()
        grid = QGridLayout(self)
        grid.setSpacing(20)
        grid.setColumnStretch(1, 10)

        l = QLabel(u"<b>Ingresá tus datos del portal Conectate:</b>")
        l.setTextFormat(Qt.RichText)
        grid.addWidget(l, 0, 0, 1, 2)

        grid.addWidget(QLabel(u"Usuario:"), 1, 0, 2, 1)
        prv = config.get('user', '')
        self.user_entry = QLineEdit(prv)
        self.user_entry.setPlaceholderText(u'Ingresá tu usuario de Conectate')
        grid.addWidget(self.user_entry, 1, 1, 2, 2)

        grid.addWidget(QLabel(u"Contraseña:"), 2, 0, 3, 1)
        prv = config.get('password', '')
        self.password_entry = QLineEdit(prv)
        self.password_entry.setEchoMode(QLineEdit.Password)
        self.password_entry.setPlaceholderText(
            u'Ingresá tu contraseña de Conectate')
        grid.addWidget(self.password_entry, 2, 1, 3, 2)

        self.password_mask = QCheckBox(u'Mostrar contraseña')
        self.password_mask.stateChanged.connect(self._toggle_password_mask)
        grid.addWidget(self.password_mask, 3, 1, 4, 2)

        l = QLabel(u'Si no tenés estos datos, '
                   u'<a href="{}">registrate aquí</a>'.format(URL_CONECTATE))
        l.setAlignment(Qt.AlignRight | Qt.AlignVCenter)
        l.setTextFormat(Qt.RichText)
        l.setOpenExternalLinks(True)
        grid.addWidget(l, 4, 0, 5, 3)

    def _toggle_password_mask(self):
        """Toggle the password hiding."""
        if self.password_mask.isChecked() is True:
            self.password_entry.setEchoMode(QLineEdit.Normal)
        else:
            self.password_entry.setEchoMode(QLineEdit.Password)

    def get_config(self):
        """Return the config for this tab."""
        d = {}
        d['user'] = self.user_entry.text()
        d['password'] = self.password_entry.text()
        return d


class PreferencesDialog(QDialog):
    """The dialog for preferences."""
    def __init__(self):
        super(PreferencesDialog, self).__init__()
        vbox = QVBoxLayout(self)

        tabbed = QTabWidget()
        self.gp = GeneralPreferences()
        tabbed.addTab(self.gp, u"General")
        self.cp = ConectatePreferences()
        tabbed.addTab(self.cp, u"Conectate")
        vbox.addWidget(tabbed)

        bbox = QDialogButtonBox(QDialogButtonBox.Ok)
        bbox.accepted.connect(self.accept)
        bbox.accepted.connect(self._save)
        vbox.addWidget(bbox)

    def closeEvent(self, event):
        """Save and close."""
        self._save()
        super(PreferencesDialog, self).closeEvent(event)

    def _save(self):
        """Just save."""
        # get it from tabs
        config.update(self.gp.get_config())
        config.update(self.cp.get_config())
        config.save()


if __name__ == '__main__':

    project_basedir = os.path.abspath(os.path.dirname(os.path.dirname(
        os.path.realpath(sys.argv[0]))))
    sys.path.insert(0, project_basedir)

    from PyQt4.QtGui import QApplication
    app = QApplication(sys.argv)

    frame = PreferencesDialog()
    frame.show()
    frame.exec_()
    frame.save_config()



#!/usr/bin/env python

# Copyright 2011-2013 Facundo Batista
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# For further info, check  https://launchpad.net/encuentro

"""Build tar.gz for encuentro.

Needed packages to run (using Debian/Ubuntu package names):

    python 2.7
    python-requests 0.12.1
    python-defer 1.0.6
    python-qt4 4.9.1
    python-xdg 0.15
    python-notify 0.1.1   # not really needed, but provides notifications
    python-bs4 4.1.0
"""

import os
import shutil

from distutils.command.install import install
from distutils.core import setup


class CustomInstall(install):
    """Custom installation class on package files.

    It copies all the files into the "PREFIX/share/PROJECTNAME" dir.
    """
    def run(self):
        """Run parent install, and then save the install dir in the script."""
        install.run(self)

        # fix installation path in the script(s)
        for script in self.distribution.scripts:
            script_path = os.path.join(self.install_scripts,
                                       os.path.basename(script))
            with open(script_path, 'rb') as fh:
                content = fh.read()
            content = content.replace('@ INSTALLED_BASE_DIR @',
                                      self._custom_data_dir)
            with open(script_path, 'wb') as fh:
                fh.write(content)

        # fix the icon path, and save the .desktop file where it should be
        src_desktop = self.distribution.get_name() + '.desktop'
        if not os.path.exists(self._custom_apps_dir):
            os.makedirs(self._custom_apps_dir)
        dst_desktop = os.path.join(self._custom_apps_dir, src_desktop)

        with open(src_desktop, 'rb') as fh:
            content = fh.read()
        icon = os.path.join(self._custom_data_dir,
                            'encuentro', 'logos', 'icon-32.png')
        content = content.replace('@ INSTALLED_ICON @', icon)
        with open(dst_desktop, 'wb') as fh:
            fh.write(content)

        # install apport file
        if not os.path.exists(self._custom_apport_dir):
            os.makedirs(self._custom_apport_dir)
        shutil.copy("source_encuentro.py", self._custom_apport_dir)

        # man directory
        if not os.path.exists(self._custom_man_dir):
            os.makedirs(self._custom_man_dir)
        shutil.copy("man/encuentro.1", self._custom_man_dir)

        # version file
        shutil.copy("version.txt", self.install_lib)

    def finalize_options(self):
        """Alter the installation path."""
        install.finalize_options(self)

        # the data path is under 'prefix'
        data_dir = os.path.join(self.prefix, "share",
                                self.distribution.get_name())
        apps_dir = os.path.join(self.prefix, "share", "applications")
        apport_dir = os.path.join(self.prefix, "share",
                                  "apport", "package-hooks")
        man_dir = os.path.join(self.prefix, "share", "man", "man1")

        # if we have 'root', put the building path also under it (used normally
        # by pbuilder)
        if self.root is None:
            build_dir = data_dir
        else:
            build_dir = os.path.join(self.root, data_dir[1:])
            apps_dir = os.path.join(self.root, apps_dir[1:])
            apport_dir = os.path.join(self.root, apport_dir[1:])
            man_dir = os.path.join(self.root, man_dir[1:])

        # change the lib install directory so all package files go inside here
        self.install_lib = build_dir

        # save this custom data dir to later change the scripts
        self._custom_data_dir = data_dir
        self._custom_apps_dir = apps_dir
        self._custom_apport_dir = apport_dir
        self._custom_man_dir = man_dir


LONG_DESCRIPTION = (
    'Simple application that allows to search, download '
    'and see the content of the Encuentro channel.'
)

setup(
    name='encuentro',
    version=open('version.txt').read().strip(),
    license='GPL-3',
    author='Facundo Batista',
    author_email='facundo@taniquetil.com.ar',
    description='Search, download and see the wonderful Encuentro content.',
    long_description=LONG_DESCRIPTION,
    url='https://launchpad.net/encuentro',

    packages=["encuentro", "encuentro.ui"],
    package_data={
        "encuentro": ["ui/media/*", "logos/icon-*.png"],
        "": ["encuentro.desktop", "source_encuentro.py", "version.txt"],
    },
    scripts=["bin/encuentro"],

    cmdclass={
        'install': CustomInstall,
    },
)



# -*- coding: utf8 -*-

# Copyright 2011-2014 Facundo Batista
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# For further info, check  https://launchpad.net/encuentro

"""Update the episodes metadata."""

import bz2
import json
import logging
from datetime import datetime
from encuentro.config import config

import defer

from encuentro import utils
from encuentro.ui import dialogs

BACKENDS_URL = "http://www.taniquetil.com.ar/encuentro/backends-v05.list"

logger = logging.getLogger('encuentro.update')


class UpdateEpisodes(object):
    """Update the episodes info."""

    def __init__(self, main_window):
        self.main_window = main_window

    def background(self):
        """Trigger an update in background."""
        self._update()

    def interactive(self):
        """Update episodes interactively."""
        dialog = dialogs.UpdateDialog()
        dialog.show()
        self._update(dialog)

    @defer.inline_callbacks
    def _update(self, dialog=None):
        """Update the content from server.

        If we have a dialog (interactive update), check frequently if
        it was closed, so we stop working for that request.
        """
        if dialog:
            tell_user = lambda *t: dialog.append(u" ".join(map(unicode, t)))
        else:
            tell_user = lambda *t: None

        logger.info("Downloading backend list")
        tell_user("Descargando la lista de backends...")
        try:
            _, backends_file = yield utils.download(BACKENDS_URL)
        except Exception, e:
            logger.error("Problem when downloading backends: %s", e)
            tell_user("Hubo un PROBLEMA al bajar la lista de backends:", e)
            return
        if dialog and dialog.closed:
            return
        backends_list = [l.strip().split() for l in backends_file.split("\n")
                         if l and l[0] != '#']

        backends = {}
        for b_name, b_dloader, b_url in backends_list:
            logger.info("Downloading backend metadata for %r", b_name)
            tell_user("Descargando la lista de episodios para backend %r..." %
                      (b_name,))
            try:
                _, compressed = yield utils.download(b_url)
            except Exception, e:
                logger.error("Problem when downloading episodes: %s", e)
                tell_user("Hubo un PROBLEMA al bajar los episodios: ", e)
                return
            if dialog and dialog.closed:
                return

            tell_user("Descomprimiendo el archivo....")
            new_content = bz2.decompress(compressed)
            logger.debug("Downloaded data decompressed ok")

            content = json.loads(new_content)
            for item in content:
                item['downtype'] = b_dloader
            backends[b_name] = content

        if dialog and dialog.closed:
            return
        tell_user("Conciliando datos de diferentes backends")
        logger.debug("Merging backends data")
        new_data = self._merge(backends)

        tell_user("Actualizando los datos internos....")
        logger.debug("Updating internal metadata (%d)", len(new_data))
        self.main_window.programs_data.merge(
            new_data, self.main_window.big_panel.episodes)

        config.update({'autorefresh_last_time': datetime.now()})
        config.save()

        tell_user(u"¡Todo terminado bien!")

        if dialog:
            dialog.accept()

    def _merge(self, backends):
        """Merge content from all backends.

        This is for v03-05, with only 'encuentro' and 'conectar' data to be
        really merged, other data just appended.
        """
        raw_encuentro_data = backends.pop('encuentro')
        raw_conectar_data = backends.pop('conectar')
        enc_data = dict((x['episode_id'], x) for x in raw_encuentro_data)
        con_data = dict((x['episode_id'], x) for x in raw_conectar_data)
        common = set(enc_data) & set(con_data)
        logger.debug("Merging: encuentro=%d conectar=%d (common=%d)",
                     len(enc_data), len(con_data), len(common))

        # what is in not common in both goes untouched
        final_data = ([enc_data[epid] for epid in set(enc_data) - common] +
                      [con_data[epid] for epid in set(con_data) - common])

        # what is common, we need to do the merge
        for epid in common:
            enc_ep = enc_data[epid]
            con_ep = con_data[epid]

            enc_desc = enc_ep['description']
            con_desc = con_ep['description']
            if enc_desc == con_desc:
                description = enc_desc
            elif enc_desc is None:
                description = con_desc
            elif con_desc is None:
                description = enc_desc
            else:
                # not None, or they would have been the same, let's concat
                # both, shorter first
                if len(con_desc) < len(enc_desc):
                    description = con_desc + ' ' + enc_desc
                else:
                    description = enc_desc + ' ' + con_desc

            # if both are equal (None or not), it also works
            if enc_ep['duration'] is None:
                duration = con_ep['duration']
            else:
                duration = enc_ep['duration']
            if enc_ep['image_url'] is None:
                image_url = con_ep['image_url']
            else:
                image_url = enc_ep['image_url']
            if enc_ep['season'] is None:
                season = con_ep['season']
            else:
                season = enc_ep['season']

            d = dict(episode_id=epid, description=description,
                     duration=duration, url=con_ep['url'],
                     channel=con_ep['channel'], title=con_ep['title'],
                     section=con_ep['section'], image_url=image_url,
                     downtype=con_ep['downtype'], season=season)
            final_data.append(d)

        logger.debug("Merging: appending other data: %s", backends.keys())
        for data in backends.itervalues():
            final_data.extend(data)
        logger.debug("Merged, final: %d", len(final_data))
        return final_data



# -*- coding: utf8 -*-

# Copyright 2013 Facundo Batista
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# For further info, check  https://launchpad.net/encuentro

"""Several dialogs."""

from PyQt4.QtGui import (
    QDialog,
    QDialogButtonBox,
    QLabel,
    QPlainTextEdit,
    QPushButton,
    QVBoxLayout,
)

UPGRADE_TEXT = u"""
Esta nueva versión del programa Encuentro sólo funciona con contenido
actualizado, lo cual le permitirá trabajar con programas del canal
Encuentro y de otros nuevos canales, pero deberá configurarlo
nuevamente y perderá la posibilidad de ver directactamente los
videos ya descargados (los cuales permanecerán en su disco).

Haga click en Continuar y podrá ver el Wizard que lo ayudará a
configurar nuevamente el programa.
"""


class ForceUpgradeDialog(QDialog):
    """The dialog for a force upgrade."""
    def __init__(self):
        super(ForceUpgradeDialog, self).__init__()
        vbox = QVBoxLayout(self)

        self.setWindowTitle(u"El contenido debe actualizarse")

        self.main_text = QLabel(UPGRADE_TEXT)
        vbox.addWidget(self.main_text)

        bbox = QDialogButtonBox()
        bbox.addButton(QPushButton(u"Salir del programa"),
                       QDialogButtonBox.AcceptRole)
        bbox.accepted.connect(self.accept)
        bbox.addButton(QPushButton(u"Continuar"),
                       QDialogButtonBox.RejectRole)
        bbox.rejected.connect(self.reject)
        vbox.addWidget(bbox)
        self.show()


class UpdateDialog(QDialog):
    """The dialog for update."""
    def __init__(self):
        super(UpdateDialog, self).__init__()
        vbox = QVBoxLayout(self)
        self.closed = False
        self.setModal(True)
        self.resize(500, 250)

        vbox.addWidget(QLabel(u"Actualización de la lista de episodios:"))
        self.text = QPlainTextEdit()
        self.text.setReadOnly(True)
        vbox.addWidget(self.text)

        bbox = QDialogButtonBox(QDialogButtonBox.Cancel)
        bbox.rejected.connect(self.reject)
        vbox.addWidget(bbox)

    def append(self, text):
        """Append some text in the dialog."""
        self.text.appendPlainText(text.strip())

    def closeEvent(self, event):
        """It was closed."""
        self.closed = True


if __name__ == '__main__':
    import sys

    from PyQt4.QtGui import QApplication
    app = QApplication(sys.argv)

    frame = UpdateDialog()
    frame.show()
    frame.exec_()



# Copyright 2013-2014 Facundo Batista
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# For further info, check  https://launchpad.net/encuentro

"""Show an icon in the systray."""

import json
import logging
import subprocess

from encuentro import multiplatform

from PyQt4.QtGui import QSystemTrayIcon, QIcon, QMenu

logger = logging.getLogger("encuentro.systray")


def _should_fix():
    """Tell if we should fix the Unity panel systray settings.

    Return None if don't need, else return the current conf.
    """
    cmd = "gsettings get com.canonical.Unity.Panel systray-whitelist".split()
    try:
        out = subprocess.check_output(cmd)
    except Exception, err:
        # don't have gsettings, nothing to fix
        etype = err.__class__.__name__
        logger.debug("No gsettings, no systray conf to fix (got %r %s)",
                     etype, err)
        return

    try:
        conf = map(str, json.loads(out.strip().replace("'", '"')))
    except ValueError:
        # don't understand the output, can't really fix it :/
        logger.warning("Don't understand gsettings output: %r", out)
        return

    logger.info("gsettings conf: %r", conf)
    if "all" in conf or "encuentro" in conf:
        # we're ok!
        return

    # need to fix
    return conf


def _fix_unity_systray():
    """Check settings."""
    conf = _should_fix()
    if conf is None:
        return

    conf.append("encuentro")
    cmd = ["gsettings", "set", "com.canonical.Unity.Panel",
           "systray-whitelist", str(conf)]
    try:
        out = subprocess.check_output(cmd)
    except OSError, err:
        logger.warning("Error trying to set the new conf: %s", err)
    else:
        logger.warning("New config set (result: %r)", out)


def show(main_window):
    """Show a system tray icon with a small icon."""
    _fix_unity_systray()
    icon = QIcon(multiplatform.get_path("encuentro/logos/icon-192.png"))
    sti = QSystemTrayIcon(icon, main_window)
    if not sti.isSystemTrayAvailable():
        logger.warning("System tray not available.")
        return

    def showhide(_):
        """Show or hide the main window."""
        if main_window.isVisible():
            main_window.hide()
        else:
            main_window.show()

    _menu = QMenu(main_window)
    _act = _menu.addAction("Mostrar/Ocultar")
    _act.triggered.connect(showhide)
    _act = _menu.addAction("Acerca de")
    _act.triggered.connect(main_window.open_about_dialog)
    _act = _menu.addAction("Salir")
    _act.triggered.connect(main_window.on_close)
    sti.setContextMenu(_menu)
    sti.show()



# Copyright 2013-2014 Facundo Batista
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# For further info, check  https://launchpad.net/encuentro

"""Get an image from web and cache it."""


import logging
import md5
import os
import glob

from encuentro import multiplatform, utils

logger = logging.getLogger('encuentro.image')


class ImageGetter(object):
    """Image downloader and cache object."""

    def __init__(self, callback):
        self.callback = callback
        self.cache_dir = os.path.join(multiplatform.cache_dir,
                                      'encuentro.images')
        if not os.path.exists(self.cache_dir):
            os.makedirs(self.cache_dir)

    def get_image(self, episode_id, url):
        """Get an image and show it using the callback."""
        logger.info("Loading image for episode %s: %r", episode_id, url)
        file_name = md5.md5(url).hexdigest()
        file_fullname = os.path.join(self.cache_dir, file_name)
        img_search_result = glob.glob(file_fullname + '.*')
        if len(img_search_result) > 0:
            logger.debug("Image already available: %r", file_fullname)
            self.callback(episode_id, file_fullname)
            return

        def _d_callback(data, episode_id, file_fullname):
            """Cache the image and use the callback."""
            content_type, img_data = data
            content, extension = content_type.split('/')
            if content != 'image':
                logger.debug("The Content-Type header is not 'image'")
            file_fullname = file_fullname + '.' + extension
            logger.debug("Image downloaded for episode_id %s, "
                         "saving to %r, Content-Type= %s",
                         episode_id, file_fullname, content_type)
            with utils.SafeSaver(file_fullname) as fh:
                fh.write(img_data)
            self.callback(episode_id, file_fullname)

        def _d_errback(failure):
            """Log the problem."""
            logger.error("Problem getting image: type: %s error: %s",
                         failure.type, failure.value)

        logger.debug("Need to download the image")
        d = utils.download(url)
        d.add_callback(_d_callback, episode_id, file_fullname)
        d.add_errback(_d_errback)
